# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from __future__ import annotations

import keystoneauth1
from magnum import objects as magnum_objects
from magnum.conductor import scale_manager
from magnum.drivers.common import driver
from magnum.objects import fields

from magnum_cluster_api import (
    clients,
    exceptions,
    monitor,
    objects,
    resources,
    sync,
    utils,
)


def cluster_lock_wrapper(func):
    def wrapper(*args, **kwargs):
        cluster = args[2]  # Assuming cluster is the second argument
        with sync.ClusterLock(cluster.uuid):
            return func(*args, **kwargs)

    return wrapper


class BaseDriver(driver.Driver):
    def __init__(self):
        self.k8s_api = clients.get_pykube_api()

    def create_cluster(
        self, context, cluster: magnum_objects.Cluster, cluster_create_timeout: int
    ):
        """
        Create cluster.

        This method is called asynchonously by the Magnum API, therefore it will not be
        blocking the Magnum API.
        """
        # NOTE(mnaser): We want to set the `stack_id` as early as possible to
        #               make sure we can use it in the cluster creation.
        cluster.stack_id = utils.generate_cluster_api_name(self.k8s_api)
        cluster.save()

        utils.validate_cluster(context, cluster)
        resources.Namespace(self.k8s_api).apply()

        return self._create_cluster(context, cluster)

    @cluster_lock_wrapper
    def _create_cluster(self, context, cluster: magnum_objects.Cluster):
        osc = clients.get_openstack_api(context)

        credential = osc.keystone().client.application_credentials.create(
            user=cluster.user_id,
            name=cluster.uuid,
            description=f"Magnum cluster ({cluster.uuid})",
        )

        resources.CloudConfigSecret(
            context,
            self.k8s_api,
            cluster,
            osc.cinder_region_name(),
            credential,
        ).apply()

        resources.ApiCertificateAuthoritySecret(context, self.k8s_api, cluster).apply()
        resources.EtcdCertificateAuthoritySecret(context, self.k8s_api, cluster).apply()
        resources.FrontProxyCertificateAuthoritySecret(
            context, self.k8s_api, cluster
        ).apply()
        resources.ServiceAccountCertificateAuthoritySecret(
            context, self.k8s_api, cluster
        ).apply()

        resources.apply_cluster_from_magnum_cluster(
            context, self.k8s_api, cluster, skip_auto_scaling_release=True
        )

    def _get_cluster_status_reason(self, capi_cluster):
        capi_cluster_status_reason = ""
        capi_ops_cluster_status_reason = ""

        # Get the latest event message of the CAPI Cluster
        capi_cluster_events = capi_cluster.events
        if capi_cluster_events:
            capi_cluster_status_reason += utils.format_event_message(
                list(capi_cluster_events)[-1]
            )

        # Get the latest event message of the CAPI OpenstackCluster
        capi_ops_cluster_events = []
        capi_ops_cluster = capi_cluster.openstack_cluster
        if capi_ops_cluster:
            capi_ops_cluster_events = capi_ops_cluster.events
        if capi_ops_cluster_events:
            capi_ops_cluster_status_reason += utils.format_event_message(
                list(capi_ops_cluster_events)[-1]
            )

        return "CAPI Cluster status: %s. CAPI OpenstackCluster status reason: %s" % (
            capi_cluster_status_reason,
            capi_ops_cluster_status_reason,
        )

    def update_cluster_control_plane_status(
        self,
        context,
        cluster: magnum_objects.Cluster,
    ):
        nodegroup = cluster.default_ng_master
        action = nodegroup.status.split("_")[0]

        kcp = resources.get_kubeadm_control_plane(self.k8s_api, cluster)
        if kcp is None:
            return nodegroup

        generation = kcp.obj.get("status", {}).get("observedGeneration", 1)
        if generation > 1:
            action = "UPDATE"

        ready = kcp.obj["status"].get("ready", False)
        failure_message = kcp.obj["status"].get("failureMessage")

        updated_replicas = kcp.obj["status"].get("updatedReplicas")
        replicas = kcp.obj["status"].get("replicas")

        if updated_replicas != replicas:
            nodegroup.status = f"{action}_IN_PROGRESS"
        elif updated_replicas == replicas and ready:
            nodegroup.status = f"{action}_COMPLETE"
        nodegroup.status_reason = failure_message

        nodegroup.save()

        return nodegroup

    @cluster_lock_wrapper
    def update_cluster_status(
        self, context, cluster: magnum_objects.Cluster, use_admin_ctx: bool = False
    ):
        # NOTE(mnaser): We may be called with a stale cluster object, so we
        #               need to refresh it to make sure we have the latest data.
        cluster.refresh()

        # TODO: watch for topology change instead
        node_groups = [
            self.update_cluster_control_plane_status(context, cluster)
        ] + self.update_nodegroups_status(context, cluster)
        osc = clients.get_openstack_api(context)

        capi_cluster = resources.Cluster(context, self.k8s_api, cluster).get_or_none()

        if cluster.status in (
            fields.ClusterStatus.CREATE_IN_PROGRESS,
            fields.ClusterStatus.UPDATE_IN_PROGRESS,
        ):
            # NOTE(mnaser): It's possible we run a cluster status update before
            #               the cluster is created. In that case, we don't want
            #               to update the cluster status.
            if capi_cluster is None:
                return

            capi_cluster.reload()
            status_map = {
                c["type"]: c["status"] for c in capi_cluster.obj["status"]["conditions"]
            }

            for condition in ("ControlPlaneReady", "InfrastructureReady", "Ready"):
                if status_map.get(condition) != "True":
                    cluster.status_reason = self._get_cluster_status_reason(
                        capi_cluster
                    )
                    cluster.save()
                    return

            api_endpoint = capi_cluster.obj["spec"]["controlPlaneEndpoint"]
            cluster.api_address = (
                f"https://{api_endpoint['host']}:{api_endpoint['port']}"
            )
            cluster.coe_version = capi_cluster.obj["spec"]["topology"]["version"]

            # NOTE(oleks): To avoid autoscaler crashes, we deploy it after the
            #              cluster api endpoint is reachable.
            if (
                cluster.status == fields.ClusterStatus.CREATE_IN_PROGRESS
                and utils.get_auto_scaling_enabled(cluster)
            ):
                resources.ClusterAutoscalerHelmRelease(self.k8s_api, cluster).apply()

            for ng in node_groups:
                if not ng.status.endswith("_COMPLETE"):
                    return
                if ng.status == fields.ClusterStatus.DELETE_COMPLETE:
                    ng.destroy()

            if cluster.status == fields.ClusterStatus.CREATE_IN_PROGRESS:
                cluster.status_reason = None
                cluster.status = fields.ClusterStatus.CREATE_COMPLETE
            if cluster.status == fields.ClusterStatus.UPDATE_IN_PROGRESS:
                cluster.status_reason = None
                cluster.status = fields.ClusterStatus.UPDATE_COMPLETE

            cluster.save()

        if cluster.status == fields.ClusterStatus.DELETE_IN_PROGRESS:
            if capi_cluster and capi_cluster.exists():
                cluster.status_reason = self._get_cluster_status_reason(capi_cluster)
                cluster.save()
                return

            # NOTE(mnaser): We delete the application credentials at this stage
            #               to make sure CAPI doesn't lose access to OpenStack.
            try:
                osc.keystone().client.application_credentials.find(
                    name=cluster.uuid,
                    user=cluster.user_id,
                ).delete()
            except keystoneauth1.exceptions.http.NotFound:
                pass

            resources.CloudConfigSecret(context, self.k8s_api, cluster).delete()
            resources.ApiCertificateAuthoritySecret(
                context, self.k8s_api, cluster
            ).delete()
            resources.EtcdCertificateAuthoritySecret(
                context, self.k8s_api, cluster
            ).delete()
            resources.FrontProxyCertificateAuthoritySecret(
                context, self.k8s_api, cluster
            ).delete()
            resources.ServiceAccountCertificateAuthoritySecret(
                context, self.k8s_api, cluster
            ).delete()

            cluster.status_reason = None
            cluster.status = fields.ClusterStatus.DELETE_COMPLETE
            cluster.save()

    @cluster_lock_wrapper
    def update_cluster(
        self,
        context,
        cluster: magnum_objects.Cluster,
        scale_manager=None,
        rollback=False,
    ):
        """
        Update cluster.

        This method is called asynchonously by the Magnum API, therefore it will not be
        blocking the Magnum API.
        """
        raise NotImplementedError()

    @cluster_lock_wrapper
    def resize_cluster(
        self,
        context,
        cluster: magnum_objects.Cluster,
        resize_manager: scale_manager.ScaleManager,
        node_count: int,
        nodes_to_remove: list[str],
        nodegroup: magnum_objects.NodeGroup = None,
    ):
        """
        Resize cluster (primarily add or remove nodes).

        The cluster object passed to this method is already not in `UPDATE_IN_PROGRESS`
        state and the node group object passed to this method is in `UPDATE_IN_PROGRESS`
        state and saved.

        This method is called asynchonously by the Magnum API, therefore it will not be
        blocking the Magnum API.
        """
        utils.validate_cluster(context, cluster)

        if nodes_to_remove:
            machines = objects.Machine.objects(self.k8s_api).filter(
                namespace="magnum-system",
                selector={
                    "cluster.x-k8s.io/cluster-name": cluster.stack_id,
                    "topology.cluster.x-k8s.io/deployment-name": nodegroup.name,
                },
            )

            for machine in machines:
                instance_uuid = machine.obj["spec"]["providerID"].split("/")[-1]
                if instance_uuid in nodes_to_remove:
                    machine.obj["metadata"].setdefault("annotations", {})
                    machine.obj["metadata"]["annotations"][
                        "cluster.x-k8s.io/delete-machine"
                    ] = "yes"
                    utils.kube_apply_patch(machine)

        self._update_nodegroup(context, cluster, nodegroup)

    @cluster_lock_wrapper
    def upgrade_cluster(
        self,
        context,
        cluster: magnum_objects.Cluster,
        cluster_template: magnum_objects.ClusterTemplate,
        max_batch_size,
        nodegroup: magnum_objects.NodeGroup,
        scale_manager=None,
        rollback=False,
    ):
        """
        Upgrade a cluster to a new version of Kubernetes.

        The only label that we change during the upgrade is the `kube_tag` label.

        Historically, the upgrade cluster has been a "hammer" that was used to sync the
        Kubernetes Cluster API objects with the Magnum objects.  However, by doing this,
        we're losing the ability to maintain the existing labels of the cluster.

        For now, upgrade cluster simply modifies the labels that are necessary for the
        upgrade, nothing else.  For the future, we can perhaps use the `update_cluster`
        API.

        This method is called synchonously by the Magnum API, therefore it will be blocking
        the Magnum API, so it should be as fast as possible.
        """
        # XXX(mnaser): The Magnum API historically only did upgrade one node group at a
        #              time.  This is a limitation of the Magnum API and not the Magnum
        #              Cluster API since doing multiple rolling upgrades was not very
        #              well supported in the past.
        #
        #              The Magnum Cluster API does not have this limitation in this case
        #              we ignore the `nodegroup` parameter and upgrade the entire cluster
        #              at once.
        cluster.cluster_template_id = cluster_template.uuid
        cluster.labels["kube_tag"] = cluster_template.labels["kube_tag"]

        for ng in cluster.nodegroups:
            ng.status = fields.ClusterStatus.UPDATE_IN_PROGRESS
            ng.image_id = cluster_template.image_id
            ng.labels["kube_tag"] = cluster_template.labels["kube_tag"]
            ng.save()

        # NOTE(mnaser): We run a full apply on the cluster regardless of the changes, since
        #               the expectation is that running an upgrade operation will change
        #               the cluster in some way.
        resources.apply_cluster_from_magnum_cluster(context, self.k8s_api, cluster)

        # NOTE(mnaser): We do not save the cluster object here because the Magnum driver
        #               will save the object that it passed to us here.

    @cluster_lock_wrapper
    def delete_cluster(self, context, cluster: magnum_objects.Cluster):
        """
        Delete cluster.

        This method is called asynchonously by the Magnum API, therefore it will not be
        blocking the Magnum API.
        """
        if cluster.stack_id is None:
            return
        # NOTE(mnaser): This should be removed when this is fixed:
        #
        #               https://github.com/kubernetes-sigs/cluster-api-provider-openstack/issues/842
        #               https://github.com/kubernetes-sigs/cluster-api-provider-openstack/pull/990
        utils.delete_loadbalancers(context, cluster)

        resources.ClusterResourceSet(self.k8s_api, cluster).delete()
        resources.ClusterResourcesConfigMap(context, self.k8s_api, cluster).delete()
        resources.Cluster(context, self.k8s_api, cluster).delete()
        resources.ClusterAutoscalerHelmRelease(self.k8s_api, cluster).delete()

    @cluster_lock_wrapper
    def create_nodegroup(
        self,
        context,
        cluster: magnum_objects.Cluster,
        nodegroup: magnum_objects.NodeGroup,
    ):
        """
        Create node group.

        The cluster object passed to this method is already in `UPDATE_IN_PROGRESS` state
        and the node group object passed to this method is in `CREATE_IN_PROGRESS` state.

        This method is called asynchonously by the Magnum API, therefore it will not be
        blocking the Magnum API.
        """
        utils.validate_nodegroup(nodegroup, context)

        cluster_resource = objects.Cluster.for_magnum_cluster(self.k8s_api, cluster)
        cluster_resource.obj["spec"]["topology"]["workers"][
            "machineDeployments"
        ].append(resources.mutate_machine_deployment(context, cluster, nodegroup))

        utils.kube_apply_patch(cluster_resource)

    def update_nodegroups_status(
        self, context, cluster: magnum_objects.Cluster
    ) -> list[magnum_objects.NodeGroup]:
        cluster_resource = objects.Cluster.for_magnum_cluster_or_none(
            self.k8s_api, cluster
        )
        if cluster_resource is None:
            return cluster.nodegroups

        node_groups = []
        for node_group in cluster.nodegroups:
            if node_group.role == "master":
                continue

            md = objects.MachineDeployment.for_node_group_or_none(
                self.k8s_api, cluster, node_group
            )
            md_is_running = (
                md is not None and md.obj.get("status", {}).get("phase") == "Running"
            )

            # NOTE(mnaser): If the node group is in `CREATE_IN_PROGRESS` state, we need to
            #               wait for the `MachineDeployment` to be hit the `Running` phase
            #               before we can mark the node group as `CREATE_COMPLETE`.
            if (
                node_group.status == fields.ClusterStatus.CREATE_IN_PROGRESS
                and md_is_running
            ):
                node_group.status = fields.ClusterStatus.CREATE_COMPLETE
                node_group.save()

            # Get list of all of the OpenStackMachine objects for this node group
            machines = objects.OpenStackMachine.objects(
                self.k8s_api, namespace="magnum-system"
            ).filter(
                selector={
                    "cluster.x-k8s.io/cluster-name": cluster.stack_id,
                    "topology.cluster.x-k8s.io/deployment-name": node_group.name,
                },
            )

            # Ensure that the image ID from the spec matches all of the OpenStackMachine objects
            # for this node group
            md_spec = cluster_resource.get_machine_deployment_spec(node_group.name)
            md_variables = {
                i["name"]: i["value"] for i in md_spec["variables"]["overrides"]
            }
            image_id_match = all(
                [
                    machine.obj["spec"]["imageUUID"] == md_variables["imageUUID"]
                    for machine in machines
                ]
            )

            # NOTE(mnaser): If the cluster is in `UPDATE_IN_PROGRESS` state, we need to
            #               wait for the `MachineDeployment` to match the desired state
            #               from the `Cluster` resource and that it is in the `Running`
            #               phase before we can mark the node group as `UPDATE_COMPLETE`.
            if (
                node_group.status == fields.ClusterStatus.UPDATE_IN_PROGRESS
                and md_is_running
                and md.equals_spec(
                    cluster_resource.get_machine_deployment_spec(node_group.name)
                )
                and image_id_match
            ):
                node_group.status = fields.ClusterStatus.UPDATE_COMPLETE
                node_group.save()

            # NOTE(mnaser): If the cluster is in `DELETE_IN_PROGRESS` state, we need to
            #               wait for the `MachineDeployment` to be deleted before we can
            #               mark the node group as `DELETE_COMPLETE`.
            if (
                node_group.status == fields.ClusterStatus.DELETE_IN_PROGRESS
                and md is None
            ):
                node_group.status = fields.ClusterStatus.DELETE_COMPLETE
                node_group.save()

            node_groups.append(node_group)

        return node_groups

    @cluster_lock_wrapper
    def update_nodegroup(
        self,
        context,
        cluster: magnum_objects.Cluster,
        nodegroup: magnum_objects.NodeGroup,
    ):
        """
        Update node group (primarily resize it)

        This cluster object passed to this method is already in `UPDATE_IN_PROGRESS` state
        and the node group object passed to this method is in `UPDATE_IN_PROGRESS` state
        but it's not saved.

        This method is called asynchonously by the Magnum API, therefore it will not be
        blocking the Magnum API.
        """
        self._update_nodegroup(context, cluster, nodegroup)

    def _update_nodegroup(
        self,
        context,
        cluster: magnum_objects.Cluster,
        nodegroup: magnum_objects.NodeGroup,
    ):
        utils.validate_nodegroup(nodegroup, context)

        cluster_resource = objects.Cluster.for_magnum_cluster(self.k8s_api, cluster)

        current_md_spec = cluster_resource.get_machine_deployment_spec(nodegroup.name)
        target_md_spec = resources.mutate_machine_deployment(
            context,
            cluster,
            nodegroup,
            cluster_resource.get_machine_deployment_spec(nodegroup.name),
        )

        if current_md_spec == target_md_spec:
            return

        cluster_resource.set_machine_deployment_spec(nodegroup.name, target_md_spec)
        utils.kube_apply_patch(cluster_resource)

        nodegroup.status = fields.ClusterStatus.UPDATE_IN_PROGRESS
        nodegroup.save()

        cluster.status = fields.ClusterStatus.UPDATE_IN_PROGRESS
        cluster.save()

    @cluster_lock_wrapper
    def delete_nodegroup(
        self,
        context,
        cluster: magnum_objects.Cluster,
        nodegroup: magnum_objects.NodeGroup,
    ):
        """
        Delete node group.

        The cluster resource that is passed to this method is already in `UPDATE_IN_PROGRESS`
        however the node group object passed to this method is in `DELETE_IN_PROGRESS` state
        but it's not saved.

        This method is called asynchonously by the Magnum API, therefore it will not be
        blocking the Magnum API.
        """

        # NOTE(mnaser): We want to switch the node group to `DELETE_IN_PROGRESS` state
        #               as soon as possible to make sure that the Magnum API knows that
        #               the node group is being deleted.
        nodegroup.status = fields.ClusterStatus.DELETE_IN_PROGRESS
        nodegroup.save()

        cluster_resource = objects.Cluster.for_magnum_cluster(self.k8s_api, cluster)

        try:
            md_index = cluster_resource.get_machine_deployment_index(nodegroup.name)
        except exceptions.MachineDeploymentNotFound:
            nodegroup.status = fields.ClusterStatus.DELETE_COMPLETE
            nodegroup.save()
            return

        del cluster_resource.obj["spec"]["topology"]["workers"]["machineDeployments"][
            md_index
        ]

        utils.kube_apply_patch(cluster_resource)

    @cluster_lock_wrapper
    def get_monitor(self, context, cluster: magnum_objects.Cluster):
        return monitor.Monitor(context, cluster)

    # def rotate_ca_certificate(self, context, cluster):
    #     raise exception.NotSupported(
    #         "'rotate_ca_certificate' is not supported by this driver.")

    def create_federation(self, context, federation):
        raise NotImplementedError()

    def update_federation(self, context, federation):
        raise NotImplementedError()

    def delete_federation(self, context, federation):
        raise NotImplementedError()


class UbuntuDriver(BaseDriver):
    @property
    def provides(self):
        return [
            {"server_type": "vm", "os": "ubuntu", "coe": "kubernetes"},
        ]


class UbuntuFocalDriver(UbuntuDriver):
    @property
    def provides(self):
        return [
            {"server_type": "vm", "os": "ubuntu-focal", "coe": "kubernetes"},
        ]


class FlatcarDriver(BaseDriver):
    @property
    def provides(self):
        return [
            {"server_type": "vm", "os": "flatcar", "coe": "kubernetes"},
        ]


class RockyLinuxDriver(BaseDriver):
    @property
    def provides(self):
        return [
            {"server_type": "vm", "os": "rockylinux", "coe": "kubernetes"},
        ]



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import itertools

import yaml


def update_manifest_images(cluster_uuid: str, file, repository=None, replacements=[]):
    with open(file) as fd:
        data = fd.read()

    docs = []
    for doc in yaml.safe_load_all(data):
        # Fix container image paths
        if doc["kind"] in ("DaemonSet", "Deployment", "StatefulSet"):
            for container in itertools.chain(
                doc["spec"]["template"]["spec"].get("initContainers", []),
                doc["spec"]["template"]["spec"]["containers"],
            ):
                for src, dst in replacements:
                    container["image"] = container["image"].replace(src, dst)
                if repository:
                    container["image"] = get_image(container["image"], repository)

        # Fix CCM cluster-name
        if (
            doc["kind"] == "DaemonSet"
            and doc["metadata"]["name"] == "openstack-cloud-controller-manager"
        ):
            for env in doc["spec"]["template"]["spec"]["containers"][0]["env"]:
                if env["name"] == "CLUSTER_NAME":
                    env["value"] = cluster_uuid

        docs.append(doc)

    return yaml.safe_dump_all(docs, default_flow_style=False)


def get_image(name: str, repository: str = None):
    """
    Get the image name from the target registry given a full image name.
    """

    if not repository:
        return name

    new_image_name = name
    if name.startswith("docker.io/calico"):
        new_image_name = name.replace("docker.io/calico/", f"{repository}/calico-")
    if name.startswith("docker.io/k8scloudprovider"):
        new_image_name = name.replace("docker.io/k8scloudprovider", repository)
    if name.startswith("registry.k8s.io/sig-storage"):
        new_image_name = name.replace("registry.k8s.io/sig-storage", repository)
    if name.startswith("registry.k8s.io/provider-os"):
        new_image_name = name.replace("registry.k8s.io/provider-os", repository)
    if name.startswith("gcr.io/k8s-staging-sig-storage"):
        new_image_name = name.replace("gcr.io/k8s-staging-sig-storage", repository)
    if new_image_name.startswith(f"{repository}/livenessprobe"):
        return new_image_name.replace("livenessprobe", "csi-livenessprobe")
    if new_image_name.startswith("registry.k8s.io/coredns"):
        return new_image_name.replace("registry.k8s.io/coredns", repository)
    if new_image_name.startswith("registry.k8s.io/autoscaling"):
        return new_image_name.replace("registry.k8s.io/autoscaling", repository)
    if (
        new_image_name.startswith("registry.k8s.io/etcd")
        or new_image_name.startswith("registry.k8s.io/kube-")
        or new_image_name.startswith("registry.k8s.io/pause")
    ):
        return new_image_name.replace("registry.k8s.io", repository)

    assert new_image_name.startswith(repository) is True
    return new_image_name



# Copyright (c) 2024 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import TestCase, mock

from oslo_utils import uuidutils

from magnum_cluster_api.sync import ClusterLock


@mock.patch("kubernetes.config.load_config")
class ClusterLockTestCase(TestCase):
    def test_cluster_lock_init_with_no_expire(self, mock_load_config):
        cluster_id = uuidutils.generate_uuid()

        lock = ClusterLock(cluster_id)

        self.assertEqual(lock.lock_name, "cluster-%s" % cluster_id)
        self.assertEqual(lock.k8s_namespace, "magnum-system")
        self.assertEqual(lock.expire, ClusterLock.DEFAULT_EXPIRE)

    def test_cluster_lock_init_with_expire(self, mock_load_config):
        cluster_id = uuidutils.generate_uuid()
        expire = 60

        lock = ClusterLock(cluster_id, expire)

        self.assertEqual(lock.lock_name, "cluster-%s" % cluster_id)
        self.assertEqual(lock.k8s_namespace, "magnum-system")
        self.assertEqual(lock.expire, expire)



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from magnum.common import exception
from magnum.i18n import _


class HelmException(Exception):
    pass


class HelmReleaseNotFound(HelmException):
    pass


class OpenStackClusterException(Exception):
    pass


class OpenStackClusterNotCreated(OpenStackClusterException):
    pass


class OpenStackClusterNotReady(OpenStackClusterException):
    pass


class OpenStackClusterExternalNetworkNotReady(OpenStackClusterNotReady):
    pass


class OpenStackClusterNetworkNotReady(OpenStackClusterNotReady):
    pass


class OpenStackClusterSubnetNotReady(OpenStackClusterNotReady):
    pass


class ClusterAPIReconcileTimeout(Exception):
    pass


class ClusterMasterCountEven(Exception):
    pass


class OpenstackFlavorInvalidName(exception.InvalidName):
    message = _("Expected a flavor name but received flavor id %(flavor)s.")


class MachineDeploymentNotFound(exception.ObjectNotFound):
    message = _("MachineDeployment %(name)s not found.")


class InvalidOctaviaLoadBalancerAlgorithm(exception.Invalid):
    message = _("Invalid value for octavia_lb_algorithm: %(octavia_lb_algorithm)s.")



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_context import context
from oslo_service import service

CONF = cfg.CONF


class Service(service.Service):
    def __init__(self, manager):
        super(Service, self).__init__()
        self.manager = manager()

    def start(self):
        self.tg.add_dynamic_timer(self.periodic_tasks)

    def periodic_tasks(self, raise_on_error=False):
        ctxt = context.get_admin_context()
        return self.manager.periodic_tasks(ctxt, raise_on_error=raise_on_error)


# NOTE: the global launcher is to maintain the existing
#       functionality of calling service.serve +
#       service.wait
_launcher = None


def serve(server, workers=None):
    global _launcher
    if _launcher:
        raise RuntimeError("serve() can only be called once")

    _launcher = service.launch(CONF, server, workers=workers, restart_method="mutate")


def wait():
    _launcher.wait()



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import pytest
import yaml
from oslo_concurrency import processutils

from magnum_cluster_api import exceptions, helm


def test_helm_upgrade(mocker):
    namespace = "test-namespace"
    release_name = "test-release"
    chart_ref = "test-chart"
    values = {"test": "value"}

    mock_execute = mocker.patch("oslo_concurrency.processutils.execute")
    mock_execute.return_value = ("", "")
    upgrade = helm.UpgradeReleaseCommand(
        namespace,
        release_name,
        chart_ref,
        values,
    )
    upgrade()

    mock_execute.assert_has_calls(
        [
            mocker.call(
                "helm",
                "upgrade",
                "--namespace",
                namespace,
                release_name,
                chart_ref,
                "--install",
                "--values",
                "-",
                process_input=yaml.dump(values),
            ),
        ]
    )


def test_helm_upgrade_with_in_progress_operation(mocker):
    namespace = "test-namespace"
    release_name = "test-release"
    chart_ref = "test-chart"
    values = {"test": "value"}

    mock_execute = mocker.patch("oslo_concurrency.processutils.execute")
    mock_execute.side_effect = processutils.ProcessExecutionError(
        stderr="Error: UPGRADE FAILED: another operation (install/upgrade/rollback) is in progress"
    )
    upgrade = helm.UpgradeReleaseCommand(
        namespace,
        release_name,
        chart_ref,
        values,
    )
    upgrade()

    mock_execute.assert_has_calls(
        [
            mocker.call(
                "helm",
                "upgrade",
                "--namespace",
                namespace,
                release_name,
                chart_ref,
                "--install",
                "--values",
                "-",
                process_input=yaml.dump(values),
            ),
        ]
    )


def test_helm_upgrade_with_existing_release(mocker):
    namespace = "test-namespace"
    release_name = "test-release"
    chart_ref = "test-chart"
    values = {"test": "value"}

    mock_execute = mocker.patch("oslo_concurrency.processutils.execute")
    mock_execute.side_effect = processutils.ProcessExecutionError(
        stderr="Error: UPGRADE FAILED: release: already exists"
    )
    upgrade = helm.UpgradeReleaseCommand(
        namespace,
        release_name,
        chart_ref,
        values,
    )
    upgrade()

    mock_execute.assert_has_calls(
        [
            mocker.call(
                "helm",
                "upgrade",
                "--namespace",
                namespace,
                release_name,
                chart_ref,
                "--install",
                "--values",
                "-",
                process_input=yaml.dump(values),
            ),
        ]
    )


def test_helm_upgrade_with_unknown_error(mocker):
    namespace = "test-namespace"
    release_name = "test-release"
    chart_ref = "test-chart"
    values = {"test": "value"}

    mock_execute = mocker.patch("oslo_concurrency.processutils.execute")
    mock_execute.side_effect = processutils.ProcessExecutionError(
        stderr="Error: UPGRADE FAILED: test-release has no deployed releases"
    )
    upgrade = helm.UpgradeReleaseCommand(
        namespace,
        release_name,
        chart_ref,
        values,
    )

    with pytest.raises(processutils.ProcessExecutionError):
        upgrade()

    mock_execute.assert_has_calls(
        [
            mocker.call(
                "helm",
                "upgrade",
                "--namespace",
                namespace,
                release_name,
                chart_ref,
                "--install",
                "--values",
                "-",
                process_input=yaml.dump(values),
            ),
        ]
    )


def test_helm_delete(mocker):
    namespace = "test-namespace"
    release_name = "test-release"

    mock_execute = mocker.patch("oslo_concurrency.processutils.execute")
    delete = helm.DeleteReleaseCommand(namespace, release_name)
    delete()

    mock_execute.assert_called_once_with(
        "helm",
        "delete",
        "--namespace",
        namespace,
        release_name,
    )


def test_helm_delete_with_no_release(mocker):
    namespace = "test-namespace"
    release_name = "test-release"

    mock_execute = mocker.patch("oslo_concurrency.processutils.execute")
    mock_execute.side_effect = processutils.ProcessExecutionError(
        stderr=f"Error: uninstall: Release not loaded: {release_name}: release: not found"
    )

    delete = helm.DeleteReleaseCommand(namespace, release_name)

    with pytest.raises(exceptions.HelmReleaseNotFound):
        delete()

    mock_execute.assert_called_once_with(
        "helm",
        "delete",
        "--namespace",
        namespace,
        release_name,
    )


def test_helm_delete_skip_missing_and_existing_release(mocker):
    namespace = "test-namespace"
    release_name = "test-release"

    mock_execute = mocker.patch("oslo_concurrency.processutils.execute")
    delete = helm.DeleteReleaseCommand(namespace, release_name, skip_missing=True)
    delete()

    mock_execute.assert_called_once_with(
        "helm",
        "delete",
        "--namespace",
        namespace,
        release_name,
    )


def test_helm_delete_with_skip_missing_and_no_release(mocker):
    namespace = "test-namespace"
    release_name = "test-release"

    mock_execute = mocker.patch("oslo_concurrency.processutils.execute")
    mock_execute.side_effect = processutils.ProcessExecutionError(
        stderr=f"Error: uninstall: Release not loaded: {release_name}: release: not found"
    )

    delete = helm.DeleteReleaseCommand(namespace, release_name, skip_missing=True)
    delete()

    mock_execute.assert_called_once_with(
        "helm",
        "delete",
        "--namespace",
        namespace,
        release_name,
    )



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from magnum import objects
from magnum.common import exception
from oslo_config import cfg

from magnum_cluster_api import clients
from magnum_cluster_api.integrations import common

CONF = cfg.CONF


def is_enabled(cluster: objects.Cluster) -> bool:
    return common.is_enabled(cluster, "cinder_csi_enabled", "volumev3")


def get_image(cluster: objects.Cluster) -> str:
    return common.get_cloud_provider_image(
        cluster, "cinder_csi_plugin_tag", "cinder-csi-plugin"
    )


def get_default_boot_volume_type(context):
    """
    Get the default boot volume type since the existing function
    magnum.common.cinder.get_default_boot_volume_type() returns a random volume
    type when CONF.cinder.default_boot_volume_type is not defined.

    Instead of using a random volume type, this function uses the default
    volume type.
    """

    if CONF.cinder.default_boot_volume_type:
        return CONF.cinder.default_boot_volume_type

    osc = clients.get_openstack_api(context)
    default_volume_type = osc.cinder().volume_types.default()

    if default_volume_type is None:
        raise exception.VolumeTypeNotFound()

    return default_volume_type.name
